/* $Workfile: T1DlBit.c $       $Revision: 3 $      $Modtime: 4/02/13 5:04p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T1RobBit.c

    A simple example console application demonstrating the use of OTX Driver
    to access DL bits with following boards:

    OTX PCI Plus series boards,

      Thor-2-PCI-Plus, Thor-4-PCI-Plus and Thor-8-PCI-Plus 

    or OTX PCI Express series boards,

      Thor-2-PCI-Express
      Thor-8-PCIe

    or Thor-2-PCMCIA series boards,

      Thor-2-PCMCIA, 
      Thor-2-PCMCIA-Plus, 
      Thor-2-PCMCIA-Ex, 
      Thor-2-PCMCIA-Pro,
      Thor-2-PCMCIA-CST

    or Sleipnir-1-PCI-Plus

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 T1 Lines (8 for Thor-8-PCI) and expects them to be looped 
    with an external loop cable.

    Keystroke command are also used to set the DL-bits to be transmitted. Please
    note that is not a typical implementation for a DL-channel protocol, but merely
    demonstrates how to use the OTX functions to transmit and read DL-bits.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"


#define DEMO_DEFAULT_LOG_FILE "T1DlBit.log"
#define DEMO_DEFAULT_BOARD_NO 0

typedef struct DemoEnvT {
  char            *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OTX_INT32       nBoardType;
  int             coLi;
  int             boardNo;
} DemoEnvT;

/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(const char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;

struct PhysDevs {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2];        // Used for Thor-4-PCI-Plus and Thor-8-PCI(-Plus)
  OTX_HANDLE hLi[8];
} g_hPhysDevs;


struct LogicDevs {
  OTX_HANDLE hT1E1DlBitAccess[8];

} g_hLogicDevs;


struct T1E1CfgData {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(/*128*/16, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI(-Plus)

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    g_demoEnv.coLi = 4;
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening Thor-2-PCMCIA variant board\n");
    g_demoEnv.coLi = 2;
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_THOR_8_PCIE:
    DemoLogf("Opening Thor-8-PCIe board\n");
    g_demoEnv.coLi = 8;
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    DemoLogf("Opening Sleipnir-1-PCI board\n");
    g_demoEnv.coLi = 1;
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    DemoLogf("Opening Thor-2-ExpressCard board\n");
    g_demoEnv.coLi = 2;
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(result);

  // Get handles to T1/E1 line interface devices 
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    // On these boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    {
      // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. 
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
      }
    }
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCIE:
    // On the Thor-8-PCI boards the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    // On the Thor-2-ExpressCard board the E1T1 devices are hosted by a dual E1T1 device. 
    DemoLogf("Open OTX_DEVICE_DUAL_T1E1\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DUAL_T1E1, 0, 0, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[0]); DemoCheckRc(result);

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1)\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[0], OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);

    // Since the demo application assumes that Li0 is looped to Li1, enable Automatic CRC bit inversion for one of the Line Intefaces (not both)
    if (i == 0) {
      g_T1E1CfgData.t1CfgData.bAutoCrcBitInversion = OTX_FALSE;
    } else {
      g_T1E1CfgData.t1CfgData.bAutoCrcBitInversion = OTX_TRUE;
    }

    result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);

    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_T1); DemoCheckRc(result);

  }

}



/*******************************************************************************
  @func
    Create the needed Logical Devices
  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Create DL-Bit Access logical devices hosted in the physical Line Interface devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hLi[i], OTX_LDEVICE_T1_DL_BIT_ACCESS, g_hEventQueue, i, &g_hLogicDevs.hT1E1DlBitAccess[i]); DemoCheckRc(result);
  }


}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  // Nothing
}


/*******************************************************************************
  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.
*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  // Nothing
}


/*******************************************************************************

  @func
    Enable Configured Devices

*******************************************************************************/
void DemoEnableDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;

  // Enable Physical Devices
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

  // Certain boards have one Quad device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) ) {
      
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  }

  // If board type is Thor-8-PCI then enable the two Quad E1T1 devices
  if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCIE) || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

  // Enable Logical DL-Bit Devices
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hT1E1DlBitAccess[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }

}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_AMI;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_AMI;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_F72;
  pT1CfgData->bCRC6Enable                = OTX_FALSE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels  

  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;
}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];
  //DemoLogf("%.8X%.8X ", pEventData->m_nTimeStamp.sUint32.nHigh, pEventData->m_nTimeStamp.sUint32.nLow);

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    // Event from T1E1 physical device (Li)
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      DemoLogf("E1/T1[%d] No string for event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (pEventData->m_nParam) {
      DemoLogf("E1/T1[%d] Event: %s (0x%x)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
    } else {
      DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
    }

    break;

  case OTX_LDEVICE_T1_DL_BIT_ACCESS:
    switch (pEventData->m_nCode) {
    case OTX_T1E1_EC_DL_BITS_RECEIVED:
      {
        static OTX_UINT32 g_nLastReceivedDlBits[8];

        assert(pEventData->m_nSourceId < 8);

        // Print DL-bits if they have changed
        // Please note that this is not a typical DL-channel protocol implementation
        if (pEventData->m_nParam != g_nLastReceivedDlBits[pEventData->m_nSourceId]) {
          g_nLastReceivedDlBits[pEventData->m_nSourceId] = pEventData->m_nParam;
          DemoLogf("DL-Bit Device[%d] DL-Bits Received (0x%X)\n", pEventData->m_nSourceId, pEventData->m_nParam);
        }
      }
      break;
    case OTX_T1E1_EC_DL_BITS_SENT:
      {
        DemoLogf("DL-Bit Device[%d] DL-Bits Sent (0x%X)\n", pEventData->m_nSourceId, pEventData->m_nParam);

        // At this time it is ok to provide the driver with more DL-bits to be transmitted,
        // but we do not do that here in this simple demo application

      }
      break;
    default:
      DemoLogf("BitRob Device[%d] Unexpected Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      break;
    }
    break;


  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}


/*******************************************************************************
  @func
    Parse the program's command line arguments and calls upon
    requested actions.

    There are some optional switches that can be used following
    the driver name on the command line. Switches are are
    separated by spaces, and can appear in arbitrary order.
    Switches consist of a hyphen and a letter:
*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr               = NULL;
  envP->bExitOnError              = OTX_FALSE;
  envP->boardNo                   = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType                = OTX_DEVICE_UNKNOWN;
  envP->coLi                      = 2;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'T') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_EXPRESS;
          } else if ( *(argv[i]+2) == '4') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'M') {
            envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'W') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCIE;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
          } else if ( *(argv[i]+2) == 'C') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_EXPRESSCARD;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }

}


/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /T[T|E|4|M|P|W|S|C: Board Type used for the demo, where\n");
  DemoLogf("                    T = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                    E = Use Thor-2-PCI-Express Board\n");
  DemoLogf("                    4 = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                    M = Use Thor-2-PCMCIA Board\n");
  DemoLogf("                    P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                    W = Use Thor-8-PCIe Board\n");
  DemoLogf("                    S = Use Sleipnir-1-PCI Board\n");
  DemoLogf("                    C = Use Thor-2-ExpressCard Board\n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");

  DemoLogf("\nPress Any Key to Exit...\n"); OsaConsoleGetCh();

}


/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
static void DemoLogf(const char *formatStr, ...)
{
  va_list argList;

  va_start(argList, formatStr);
  OsaConsolePrintF((char *)formatStr, argList);
  va_end(argList);

  if (g_pDemoLogFile){
   
    va_start(argList, formatStr);
    vfprintf(g_pDemoLogFile,   formatStr, argList);
    va_end(argList);
  }

}


static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {
    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
    
    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}



/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32 nCurrentLi = 0;
  OTX_RESULT result;

  switch (commandChar) {
    
  case '?':
  case 'h':
  case 'H':
    {
      DemoLogf("Cmd  Synopsis\n");
      DemoLogf("===  ========\n");
      DemoLogf(" ?   Display Help text\n");
      DemoLogf(" h   Ditto\n");
      DemoLogf(" q   Quit\n");
      DemoLogf(" 0-7 Select current Li. (currently Li=%d selected)\n", nCurrentLi);
      DemoLogf(" l   Query Line Status of all Lis\n");
      DemoLogf(" e/E Enable/Disable the DL-Bit Access Device on current Li\n");
      DemoLogf(" d   Transmit DL-Bit data on current Li\n");
    }
    break;
    
  case 'e':
    {
      
      DemoLogf("Enable the T1/E1 DL-Bit Access Device on LI#%d \n", nCurrentLi);
      
      result = OtxDrvEnable(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi],   OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      
      // Provide the driver with initial DL-bits to be transmitted.
      // When these DL-bit have been transmitted the DL-Bit device will generate
      // a OTX_T1E1_EC_DL_BITS_SENT event. At that time we can provide more
      // DL-bits to the device to be transmitted.
      //result = OtxT1SetDlBitData(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi], 0x000000); DemoCheckRc(result);
      
    }
    break;
    
  case 'E':
    {
      
      DemoLogf("Disable the T1/E1 DL-Bit Access Device on LI#%d \n", nCurrentLi);
      
      result = OtxDrvDisable(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi],   OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
    break;
    
  case 'd':
    {
      char szBuf[12]; 
      static OTX_UINT32 nDlBits = 0;
      
      DemoLogf("Enter DL-Bit data to transmit on LI#%d [0x%X]: ", nCurrentLi, nDlBits);
      OsaConsoleGetStr(szBuf);
      if (*szBuf) {
        nDlBits = strtoul(szBuf, NULL, 0);
      }
      
      // Provide the driver with initial DL-bits to be transmitted.
      // When these DL-bit have been transmitted the DL-Bit device will generate
      // a OTX_T1E1_EC_DL_BITS_SENT event. At that time we can provide more
      // DL-bits to the device to be transmitted.
      DemoLogf("Sending DL-Bit data: 0x%X\n", nDlBits);
      result = OtxT1SetDlBitData(g_hLogicDevs.hT1E1DlBitAccess[nCurrentLi], nDlBits); DemoCheckRc(result);
      
    }
    break;
    
  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 i;
      
      for (i = 0; i < g_demoEnv.coLi; i++) {
        DemoLogf("Line status of Li #%d is ", i);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[i], &eLayer1Status); DemoCheckRc(result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;
    
    
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    {
      // Select current Li
      OTX_INT32 nLi = commandChar - '0';
      
      if (nLi > g_demoEnv.coLi) {
        DemoLogf("Error: LI#%d is not supported by this board\n", nLi);
      } else {
        nCurrentLi = nLi;
        DemoLogf("LI#%d Selected ....\n", nCurrentLi);
      }
    }
    break;
    
  case 'i':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      
      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      
      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(result);
      
      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;
    
  case 'o':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;
      
      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }
      
      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(result);
      
      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;
    
    
  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;
    
  case 'q':
  case 'Q':
    {
      OTX_INT32 i;
      // Disable Logical DL-Bit Devices
      for (i = 0; i < g_demoEnv.coLi; i++) {
        result = OtxDrvDisable(g_hLogicDevs.hT1E1DlBitAccess[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
      }
    }
    return 1;
    break;
    
  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


/*******************************************************************************

  @func

  Enumerates and finds the first board device

*******************************************************************************/
static OTX_RESULT FindBoardDevice(OTX_INT32 *peBrdType) 
{
  OTX_UINT32   eDevType;
  OTX_ITERATOR itrChildTypes;
  OTX_ITERATOR itrChildrenOfSameType;
  static char szBuf[256];
  OTX_RESULT otxRc;

  *peBrdType = OTX_DEVICE_UNKNOWN; // Defensive
  
  // Initialize the opaque iterator for the hosted types
  OtxDrvIteratorBegin(&itrChildTypes);
  while (OTX_SUCCESS(OtxDrvEnumDeviceType(0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {   // Note that we must use a zero handle to get the children of the root level
    OTX_HANDLE hChildDevice;
    // Initialize the opaque iterator for the devices within the same type
    OtxDrvIteratorBegin(&itrChildrenOfSameType);
    while (OTX_SUCCESS(OtxDrvEnumDevice(0, eDevType, &itrChildrenOfSameType, &hChildDevice))) {
      if (hChildDevice != OTX_INVALID_HANDLE_VALUE) {

        // Check device type
        switch (eDevType) {
          // These devices appear at the top (driver level) but they are not board devices, so ignore them here
        case OTX_DEVICE_OTPD_INTERFACE:
        case OTX_DEVICE_CT_BUS:
          break;
        default:
          otxRc = OtxDrvGetAttributeValueSz(hChildDevice, OTX_ATTR_DEVICE_STRING, szBuf, 256);
          if (!OTX_SUCCESS(otxRc)) {
            DemoLogf("Could not get the OTX_ATTR_DEVICE_STRING attribute for device 0x%X\n", hChildDevice);
            return otxRc;
          } 
          DemoLogf("Found %s board (devType=%d)\n", szBuf, eDevType);
          *peBrdType = eDevType;
          return OTX_S_OK;
        }

      }
    }
  }

  return OTX_S_FALSE; // We did not find an OTX board
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  DemoLogf("T1DlBit, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Example program for accessing T1 DL-bit data\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_UNKNOWN) {
    // Board type was not supplied by a command line parameter, so enumerate and find a board
    result = FindBoardDevice(&g_demoEnv.nBoardType);
    if (result == OTX_S_FALSE) {
      DemoLogf("No OTX board was found\n");
      DemoExitQuestion();
    } else if (result != OTX_S_OK) {
      DemoExitQuestion();
    }
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Set clock source for the board (important to set the clock source before configuring the E1/T1 devices)
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(result);

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable all devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableDevices();


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300);
  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

