/* $Workfile: EnumOtxDevices.c $       $Revision: 5 $      $Modtime: 4/04/13 9:19a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver Test Program

  @doc PublicDemo

  @module

    EnumOtxDevices.c

    A simple example console application demonstrating the use of OTX Driver
    with the OTX boards.

		The application enumerates board serial number and available
    physical devices.
		
    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OsaThrd.h"

#define DEMO_DEFAULT_LOG_FILE "EnumOtxDevices.log"
#define DEMO_DEFAULT_BOARD_NO 0

typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  int             boardNo;
} DemoEnvT;

static int ms_nIndentLevel = 0;

/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void ReadSerialNum(OTX_HANDLE hParentDevice, OTX_HANDLE hChildDevice, OTX_INT32 eDevType);
static void EnumerateDevices(OTX_HANDLE hParentDevice);
static const char *DemoGetDeviceTypeString(IN OTX_HANDLE eDevType);
static signed int DemoGetDeviceNo(IN OTX_HANDLE hDevice);
static void DemoLogIndent();

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(const char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;



/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {									// Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {		// Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {								// Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {					// Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(const char *formatStr, ...)
{
  va_list argList;

  va_start(argList, formatStr);
  OsaConsolePrintF((char *) formatStr, argList);
  va_end(argList);

  if (g_pDemoLogFile) {
    va_start(argList, formatStr);
    vfprintf(g_pDemoLogFile,   formatStr, argList);
    va_end(argList);
  }


}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}



/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
static OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" a   Test attributes\n");
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  case 'a':
    {
      OtxVersionInfo sVersionInfo;
      char szVersionInfo[255];
      OTX_RESULT result;

      result = OtxDrvGetAttributeValueSz(0, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(0, result);
      result = OtxDrvGetAttributeValue(0, OTX_ATTR_DRV_REV_NUM, &sVersionInfo, sizeof(OtxVersionInfo), NULL); DemoCheckRc(0, result);

      DemoLogf("Driver version string     %s\n", szVersionInfo);
      DemoLogf("Driver major version      %d\n", sVersionInfo.nMjRev);
      DemoLogf("Driver minor version      %d\n", sVersionInfo.nMnRev);
      DemoLogf("Driver status             %s\n", sVersionInfo.bRelStatus ? "Released" : "Preliminary");
      DemoLogf("Driver preliminary level  %d\n", sVersionInfo.nPrelLevel);

    }

    EnumerateDevices(0);
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************

  @func

    Read board serial number

*******************************************************************************/
static void ReadSerialNum(OTX_HANDLE hParentDevice, OTX_HANDLE hChildDevice, 
                          OTX_INT32 eDevType)
{
  OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
	OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
	OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
	OTX_RESULT      result;
	char *pszSerialNumber;
  char *pszBrdBuildNumber;
  char *pszBrdUserString;
  char *pszCpldBuildNumbers;
	OTX_UINT32      nValue;  
				
  if( hParentDevice == 0 && eDevType != -1)		// Parent device with children
	{
    result = OtxDrvGetAttributeType(hChildDevice, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(hChildDevice, result);
		assert(eAttribType == OTX_ATTRIB_SZ);
		assert(bReadOnly == OTX_TRUE);
		pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
		result = OtxDrvGetAttributeValueSz(hChildDevice, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(hChildDevice, result);
		DemoLogf("\nBoard Serial Number: %s \n", pszSerialNumber);
		free(pszSerialNumber);

    result = OtxDrvGetAttributeType(hChildDevice, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(hChildDevice, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(hChildDevice, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(hChildDevice, result);
    DemoLogf("  Board-Build: %s", pszBrdBuildNumber);
    free(pszBrdBuildNumber);

    DemoLogf("\n");

    result = OtxDrvGetAttributeType(hChildDevice, OTX_ATTR_BRD_USER_STRING, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(hChildDevice, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdUserString = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(hChildDevice, OTX_ATTR_BRD_USER_STRING, pszBrdUserString, cbAttribValue); DemoCheckRc(hChildDevice, result);
    if (*pszBrdUserString) {
      DemoLogf("  Board-UserString: %s", pszBrdUserString);
      DemoLogf("\n");
    }
    free(pszBrdUserString);

    result = OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_BRD_DEVICE_ID, &nValue); DemoCheckRc(hChildDevice, result);
    DemoLogf("  DeviceId: %d", nValue);

    result = OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_BRD_SUBSYSTEM_ID, &nValue); DemoCheckRc(hChildDevice, result);
    DemoLogf("  SubSystemId: 0x%X", nValue);

    DemoLogf("\n");

    result = OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_BRD_PCI_DEVICE_NUMBER, &nValue); 
    if (OTX_RESULT_CODE(result) == OTX_S_OK) {
      DemoLogf("  PciDevNo: 0x%X", nValue);
    }

    result = OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_BRD_PCI_FUNCTION_NUMBER, &nValue); 
    if (OTX_RESULT_CODE(result) == OTX_S_OK) {
      DemoLogf("  PciFuncNo: 0x%X\n", nValue);
    }

    result = OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_BRD_PCMCIA_SOCKET_NUMBER, &nValue); 
    if (OTX_RESULT_CODE(result) == OTX_S_OK) {
      DemoLogf("  PcmciaSocketNo: 0x%X\n", nValue);
    }


    result = OtxDrvGetAttributeType(hChildDevice, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); 
    if (OTX_RESULT_CODE(result) == OTX_S_OK) {
      assert(eAttribType == OTX_ATTRIB_SZ);
      assert(bReadOnly == OTX_TRUE);
      pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
      result = OtxDrvGetAttributeValueSz(hChildDevice, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(hChildDevice, result);
      DemoLogf("  CPLD-Build:\n%s", pszCpldBuildNumbers);
      free(pszCpldBuildNumbers);
    }

    DemoLogf("\n");
	}

}


/*******************************************************************************

  @func

  Enumerates all devices for a given parent devices and prints attributes
  etc.

  @comm

*******************************************************************************/
static void EnumerateDevices(OTX_HANDLE hParentDevice)
{
  OTX_UINT32   eDevType;
  OTX_ITERATOR itrChildTypes;
  OTX_ITERATOR itrChildrenOfSameType;

  // This is a recursive function so we indent one more level for each recursion

  // Initialize the opaque iterator for the hosted types
  OtxDrvIteratorBegin(&itrChildTypes);
  while (OTX_SUCCESS(OtxDrvEnumDeviceType(hParentDevice, &itrChildTypes, (OTX_INT32 *)&eDevType))) {   // Note that we must use a zero handle to get the children of the root level
    OTX_HANDLE hChildDevice;
    // Initialize the opaque iterator for the devices within the same type
    OtxDrvIteratorBegin(&itrChildrenOfSameType);
    while (OTX_SUCCESS(OtxDrvEnumDevice(hParentDevice, eDevType, &itrChildrenOfSameType, &hChildDevice))) {
      if (hChildDevice != OTX_INVALID_HANDLE_VALUE) {
        OTX_UINT32 eDevType2;
        OTX_HANDLE hParentDevice2;

				// Here we have a handle for a child device, get the attributes
        DemoLogIndent();
        DemoLogf("%s_%d (%s, handle = %#x)\n", 
          DemoGetDeviceTypeString(hChildDevice), 
          DemoGetDeviceNo(hChildDevice), 
          (eDevType >= 0) ? (eDevType == 0 ? "UnKnown" : "Physical") : "Logical", 
          hChildDevice);

        // Verify that the type and host attributes correspond to the ones we have since earlier
        OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_DEVICE_TYPE, &eDevType2);
        OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_DEVICE_HOST, &hParentDevice2);
        if (eDevType2 != eDevType) {
          DemoLogIndent();DemoLogf("  Incorrect attribute: OTX_ATTR_DEVICE_TYPE\n");
        }
        if (hParentDevice2 != hParentDevice) {
          DemoLogIndent();DemoLogf("  Incorrect attribute: OTX_ATTR_DEVICE_HOST\n");
        }

        // Print device specific attributes
        switch (eDevType) {
        case OTX_LDEVICE_PIPE:
          {
            OTX_UINT32 nValue;
            OTX_UINT8  sqConnMask[OTX_ANY_PIPE_TIMESLOTS_MAX];
            unsigned int i;

            OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_PIPE_CAPACITY, &nValue);
            DemoLogIndent();DemoLogf("  Capacity = %d\n", nValue);

            OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_PIPE_IN_CONNECTED, &nValue);
            if (nValue != OTX_INVALID_HANDLE_VALUE) {
              DemoLogIndent();DemoLogf("  Input connected to device with handle = %#x (%s_%d)\n", nValue, DemoGetDeviceTypeString(nValue), DemoGetDeviceNo(nValue));
              OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_PIPE_IN_HIGHWAY, &nValue);
              DemoLogIndent();DemoLogf("  Input connected to highway = %d\n", nValue);
              OtxDrvGetAttributeValue(hChildDevice, OTX_ATTR_PIPE_IN_CONN_MASK, sqConnMask, OTX_ANY_PIPE_TIMESLOTS_MAX, &nValue);
              DemoLogIndent();DemoLogf("  Input connection mask = ");
              for (i = 0; i < nValue; i++) {
                DemoLogf(" %0.2x", sqConnMask[i]);
              }
              DemoLogf("\n");

            } else {
              DemoLogIndent();DemoLogf("  Input not connected\n");
            }

            OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_PIPE_OUT_CONNECTED, &nValue);
            if (nValue != OTX_INVALID_HANDLE_VALUE) {
              
              DemoLogIndent();DemoLogf("  Output connected to device with handle = %#x (%s_%d)\n", nValue, DemoGetDeviceTypeString(nValue), DemoGetDeviceNo(nValue));
              OtxDrvGetAttributeValueUint32(hChildDevice, OTX_ATTR_PIPE_OUT_HIGHWAY, &nValue);
              DemoLogIndent();DemoLogf("  Output connected to highway = %d\n", nValue);

              OtxDrvGetAttributeValue(hChildDevice, OTX_ATTR_PIPE_OUT_CONN_MASK, sqConnMask, OTX_ANY_PIPE_TIMESLOTS_MAX, &nValue);
              DemoLogIndent();DemoLogf("  Output connection mask = ");
              for (i = 0; i < nValue; i++) {
                DemoLogf(" %0.2x", sqConnMask[i]);
              }
              DemoLogf("\n");
            } else {
              DemoLogIndent();DemoLogf("  Output not connected\n");
            }

          }
          break;
        case OTX_LDEVICE_EVENT_QUEUE:
          break;
        case OTX_DEVICE_LI_POTS:
          break;
        case OTX_DEVICE_ARNI_PCI:     
        case OTX_DEVICE_BALDER_PCMCIA:
        case OTX_DEVICE_BALDER_S_PCI:
        case OTX_DEVICE_BALDER_U_PCI:
        case OTX_DEVICE_THOR_PCMCIA:
        case OTX_DEVICE_VIDAR_5x_PCI:
        case OTX_DEVICE_VIDAR_6x_PCI:
        case OTX_DEVICE_SIF_PCI:
        case OTX_DEVICE_THOR_PCI:
        case OTX_DEVICE_THOR_8_PCI:
        case OTX_DEVICE_GIMLE_16_PCI_PLUS:
        case OTX_DEVICE_GIMLE_16_PCIE:
        case OTX_DEVICE_GIMLE_16_PCIE_BASIC_32:
        case OTX_DEVICE_GIMLE_16_PCIE_BASIC_64:
        case OTX_DEVICE_THOR_1_PCI_BASIC:
        case OTX_DEVICE_THOR_8_PCI_PLUS:       
        case OTX_DEVICE_THOR_2_PCI_PLUS:       
        case OTX_DEVICE_SLEIPNIR_1_PCI:        
        case OTX_DEVICE_VIDAR_55X12_PCI:       
        case OTX_DEVICE_THOR_4_PCI_PLUS:       
        case OTX_DEVICE_THOR_1_PCI_PLUS:       
        case OTX_DEVICE_ARNI_6X6_PCI_PLUS:     
        case OTX_DEVICE_ARNI_6X6_PCI:          
        case OTX_DEVICE_ARNI_6_FXS_PCI_PLUS:   
        case OTX_DEVICE_ARNI_6_FXO_PCI_PLUS:   
        case OTX_DEVICE_ARNI_6X3_PCI_PLUS:     
        case OTX_DEVICE_HUGIN_2_PC104:         
        case OTX_DEVICE_GIMLE_16_PCI_BASIC_32: 
        case OTX_DEVICE_GIMLE_16_PCI_BASIC_64: 
        case OTX_DEVICE_THOR_1_PCI_EXPRESS:    
        case OTX_DEVICE_THOR_2_PCI_EXPRESS:
        case OTX_DEVICE_THOR_2_EXPRESSCARD:
        case OTX_DEVICE_THOR_4_EXPRESSCARD:
        case OTX_DEVICE_THOR_8_PCIE:       
        case OTX_DEVICE_THOR_8_USB:
          {
           	OTX_RESULT      result;

            // Read serial number from the board
            ReadSerialNum( hParentDevice, hChildDevice, eDevType );

            // Just to demonstrate the use of OtxDrvOpenPhysicalDeviceEx(). For simplicity we use the DeviceNo as the nEventSourceId, and no associated event queue (OTX_INVALID_HANDLE_VALUE) in this call.
            if (eDevType) {
              DemoLogf("Opening board# %d using OtxDrvOpenPhysicalDeviceEx()\n", DemoGetDeviceNo(hChildDevice));
              
              result = OtxDrvOpenPhysicalDeviceEx(hChildDevice, DemoGetDeviceNo(hChildDevice), OTX_INVALID_HANDLE_VALUE); DemoCheckRc(hChildDevice, result);
            }
          }
          break;
        }

        // Now do the same thing for all children of this device (recursion)
        ms_nIndentLevel += 1;
        EnumerateDevices(hChildDevice);
        ms_nIndentLevel -= 1;
      } else {
        // This should not happen
        DemoLogf("Internal Error: queried a handle value but it still didn't work??? Should not happen!\n");
      }
    }
  }
}


// Returns the string representing the device type for a specific device, automatically chooses between logical and physical devices
static const char *DemoGetDeviceTypeString(IN OTX_HANDLE hDevice)
{
  static char szBuf[256];
  OTX_RESULT otxRc;
  
  otxRc = OtxDrvGetAttributeValueSz(hDevice, OTX_ATTR_DEVICE_STRING, szBuf, 256);
  if (!OTX_SUCCESS(otxRc)) {
      DemoLogf("Could not get the OTX_ATTR_DEVICE_STRING attribute for device 0x%X\n", hDevice);
      return "*** Error ***";
  } else {
    return szBuf;
  }
}


// Returns the device no for a specific device
static signed int DemoGetDeviceNo(IN OTX_HANDLE hDevice)
{
  OTX_UINT32 nDevNo;
  if (OTX_SUCCESS(OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDevNo))) {
    return nDevNo;
  } else {
    return -1;
  }
}


/*******************************************************************************

  @func

  Indents according to ms_nIndentLevel, before a DemoLogf()

  @comm

*******************************************************************************/
static void DemoLogIndent()
{
  fprintf(stdout, "%*s", ms_nIndentLevel*2, "");

  if (g_pDemoLogFile) {
    fprintf(g_pDemoLogFile, "%*s", ms_nIndentLevel*2, "");
  }
}



/*******************************************************************************
	main
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  DemoLogf("EnumOtxDevices, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("OTX Test Program.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
		OsaConsoleExit();
    exit(-1);
  }

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);

  // Connect to the Driver
  DemoLogf("Connecting lib...");
  result = OtxDrvConnectLib(); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  DemoLogf("Lib connected okay\n");

  DemoLogf("\nPress 'h' for help\n");
  DemoLogf("Press 'a' to enumerate OTX devices\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}
